import java.util.*;
import java.io.File;
import java.io.FileNotFoundException;

import javax.swing.JFileChooser;

/**
 * @author Eric Durant <durant@msoe.edu>
 * Created on October 12, 2014
 * Last updated NA
 * Java sample solutions for Op 2014
 */
public class Op2014 {
    /**
     * Console menu allowing selection of any of the 9 demo solutions
     * @param args (unused)
     */
    public static void main(String[] args)
    {
        boolean done = false;
        while(!done)
        {
            System.out.println("Op 2014 Demo Program");
            System.out.println("1. Count double letters");
            System.out.println("2. Photograph exposure time");
            System.out.println("3. One week from now");
            System.out.println("4. Median");
            System.out.println("5. Prime factor exponent knockdown");
            System.out.println("6. Car loan amortization");
            System.out.println("7. License plate game");
            System.out.println("8. Gray code");
            System.out.println("9. The friend of my friend is my enemy");
            System.out.println("0. Exit");

            int choice = -1; // invalid
            while (choice > 9 || choice < 0)
            {
                if (sc.hasNextInt()) {
                    choice = sc.nextInt();
                    sc.nextLine(); // dump CR from stream
                } else
                    sc.next(); // discard non-int token
            }

            try {
                switch(choice)
                {
                    case 1: prob1(); break;
                    case 2: prob2(); break;
                    case 3: prob3(); break;
                    case 4: prob4(); break;
                    case 5: prob5(); break;
                    case 6: prob6(); break;
                    case 7: prob7(); break;
                    case 8: prob8(); break;
                    case 9: prob9(); break;
                    case 0: done = true; break;
                }
            } catch (FileNotFoundException e) {
                System.err.println("Op Demo returned to main menu due to FileNotFound exception: " + e.getMessage());
            }
        }
    }

    /**
     * Problem 1. Count double letters (10 Points)
     */
    private static void prob1() {
        /* Write a program that requires the user to enter a word and which reports to the user how many doubled letters the word contains.
         * A doubled letter is the same letter appearing twice without an intervening character.
         * Be careful to handle upper and lowercase correctly (‘a’ and ‘A’ are the same letter).
         * Assume that the user correctly enters a word with no invalid characters such as spaces or punctuation.
         * Also assume that the word does not contain any tripled, quadrupled, etc. letters.
         */
        final String word = inputLine("Enter a word: ");
        char previous = '?'; // initially invalid
        int nDoubled = 0;
        for(char next : word.toLowerCase().toCharArray()) {
            if (next == previous) {
                ++nDoubled;
            }
            previous = next;
        }
        System.out.println("That word has " + nDoubled + " doubled letters.");
    }

    /**
     * Problem 2. Photograph exposure time (10 Points)
     */
    private static void prob2() {
        /*
         * Write a program that calculates and displays the required exposure time in seconds (shutter speed, t)
         * to take a properly exposed photograph given a measure of how bright the scene is (the exposure value, V)
         * and how much light the lens lets in (its aperture or f-number, N).
         * These quantities are all real numbers and are related by the equation V = log_2(N^2/t).
         * The f-number must be positive, so you must re-prompt the user repeatedly while they do not enter a positive value.
         * There is no restriction on V except that it be a real number. You may assume all numbers entered by the user are real.
         */
        final double V = inputDouble("EV (exposure value)? ");
        double N;
        do {
            N = inputDouble("f-number (must be positive)? ");
        } while (N <= 0);
        final double t = N * N / Math.pow(2, V);
        System.out.println("The required exposure time is " + t + " seconds.");
    }

    /**
     * Problem 3. One week from now (10 Points)
     */
    private static void prob3() {
        /*
         * Prompt the user to enter three separate integers, in order: “Year: ” (2014 must be entered as a 4-digit number, etc.),
         * “Month: ” (1 represents January, 2 February, etc.), and “Day: ” (1 through 31, although the upper limit is less for some months).
         * You may assume that all user inputs are integers. You do not need to check the validity of the year.
         * But, you must ensure that the month is a number between 1 and 12, re-prompting the user for the month as necessary.
         * And, you must ensure that the day number entered is legal given the month, re-prompting the user for the day number as needed.
         * Assume that February always has 28 days.
         * Calculate the date (year, month, and day) that is one week from the input date and output it in the format Y-M-D.
         * For example, December 5, 2014 is 2014-12-5.
         */
        int year = inputInt("Year: ");
        final int daysInMonth[] = new int[] {31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};
        int month;
        do {
            month = inputInt("Month: ");
        } while (month < 1 || month > daysInMonth.length);
        int day;
        do {
            day = inputInt("Day: ");
        } while(day < 1 || day > daysInMonth[month-1]);

        day += 7;

        if (day > daysInMonth[month-1]) {
            day -= daysInMonth[month++-1];
            if (month > daysInMonth.length) {
                month = 1;
                ++year;
            }
        }

        System.out.printf("One week from now is %d-%d-%d.\n", year, month, day);
    }

    /**
     * Problem 4. Median (20 Points)
     */
    private static void prob4() throws FileNotFoundException {
        /*
         * Given a file that contains real numbers separated by whitespace (spaces, newlines, etc.),
         * read in all the numbers and calculate and display the median.
         * You may assume that the numbers are properly formatted and that there is at least 1 number.
         * The median is defined as the middle number in a sorted list of odd length.
         * For a list of even length, the median is the average of the two middle numbers when the list is sorted.
         */

        JFileChooser chooser = new JFileChooser(System.getProperty("user.dir"));
        //chooser.setFileFilter(new FileNameExtensionFilter("Text files", "txt"));
        if(chooser.showOpenDialog(null) != JFileChooser.APPROVE_OPTION) {
            throw new IllegalArgumentException("User did not select a file");
        }
        // Open the specified file
        final Scanner fsc = new Scanner(chooser.getSelectedFile()); // throws FileNotFoundException
        final List<Double> li = new ArrayList<>();
        while (fsc.hasNextDouble()) {
            li.add(fsc.nextDouble());
        }
        Collections.sort(li);
        double median;
        final int idx = li.size()/2;
        median = li.get(idx);
        if (li.size() % 2 == 0) {
            median += li.get(idx - 1);
            median /= 2;
        }
        System.out.println("The median is " + median);
    }

    /**
     * Problem 5. Prime factor exponent knockdown (20 Points)
     */
    private static void prob5() {
        /*
         * Ask the user to enter a positive integer; you may assume that their input is valid.
         * Then, calculate and display the knocked down version of the number where the exponent of each prime factor is reduced by 1.
         * So, for inputs that have no repeated prime factors, the output will be 1.
         * More formally, find the prime factorization of the input as p_1^(n_1 ) p_2^(n_2 )⋯p_k^(n_k ) and output
         * p_1^(n_1-1) p_2^(n_2-1)⋯p_k^(n_k-1), where k is the number of unique prime factors. Examples:
         *  44,100 = 2^2 3^2 5^2 7^2 → 2•3•5•7 = 210
         *  1,960 = 2^3 5•7^2 → 2^2 5^0 7^1 = 4•1•7 = 28
         */
        final int input = inputInt("Positive integer to knock down: ");
        List<Integer> li = new ArrayList<>();
        int factor = 2;
        int remaining = input;
        while (factor * factor < remaining) {
            if (remaining % factor == 0) {
                li.add(factor);
                remaining /= factor;
            } else {
                ++factor;
            }
        }
        if (factor * factor == remaining) {
            li.add(factor);
            li.add(factor);
        } else {
            li.add(remaining);
        }
        prob5sanityCheck(input, li); // not necessary for students to do, but this checks our factoring algorithm
        int result = 1;
        int lastFactor = -1; // invalid
        for(Integer i: li) {
            if (i == lastFactor) {
                result *= lastFactor;
            } else {
                lastFactor = i;
            }
        }
        System.out.println("The knocked-down version is " + result);
    }
    private static void prob5sanityCheck(int input, List<Integer> li) {
        int prod = 1;
        for (int i: li) prod *= i;
        if (prod != input) System.err.printf("Unexpected: Error in factoring algorithm: Input=%d, ProductOfFactors=%d.\n", input, prod);
    }

    /**
     * Problem 6. Car loan amortization (20 Points)
     */
    private static void prob6() {
        /*
         * Ask the user to input the following values for a car loan:
         * * Principal amount of loan in US dollars, P. Ensure this is a positive, real number, re-prompting as necessary.
         * * Ask the user to input the following values for a car loan:
         * * Principal amount of loan in US dollars, P. Ensure this is a positive, real number, re-prompting as necessary.
         * * * You may assume the user enters a number; you may not assume it is positive.
         * * Annual interest rate in percent. Ensure this is a positive, real number, again re-prompting the user as necessary.
         * * * You may assume the user enters a number; you may not assume it is positive.
         * * * For use in the formula below, this must be converted to an unscaled, monthly rate, r.
         *     So, if the user enters “6.6”, representing 6.6% per year, you will use (6.6/100)/12 = 0.0055 as the monthly interest rate in the algorithm.
         * * Term of loan in months, N. Ensure this is a positive integer, re-prompting the user as necessary.
         * * * You may assume the user enters an integer; you may not assume it is positive.
         *
         * You’ll then calculate the monthly payment using c=rP/(1-〖(1+r)〗^(-N) )
         *
         * Then, you’ll output a table with a row for each month showing:
         * * Month number (1…N)
         * * Interest amount (r × the remaining principal of the loan at the beginning of the month)
         * * Amount toward principal (c – this month’s interest amount)
         * * Remaining principal (last month’s remaining principal – this month’s amount towards principal)
         *
         * Display all output dollar amounts to the nearest cent, e.g., $52.39.
         * You should not account for the effect of the rounding to the nearest cent. When you do everything correctly,
         * the remaining principal at the end will be $0.00 to the nearest cent.
         * (In a real loan, payments and interest are in whole cents and as a result the final payment will usually differ by many cents.)
         */
        double P = -1;
        while (P <= 0) {
            P = inputDouble("Principal (must be positive and real): ");
        }
        double r = -1;
        while (r <= 0) {
            r = inputDouble("Annual interest percentage (must be positive and real): ");
        }
        r = (r/100)/12; // convert to unscaled, monthly rate
        int N = 0;
        while (N < 1) {
            N = inputInt("Term in months (must be positive): ");
        }

        final double c = r*P / (1-Math.pow(1+r, -N));
        System.out.println("Month\tInterest\tPrincipal\tPrincipal Remain.");
        double pRem = P;
        for (int row = 1; row <= N; ++row) {
            final double iAmt = r*pRem;
            final double pAmt = c-iAmt;
            pRem -= pAmt;
            System.out.printf("%d\t$%.2f\t$%.2f\t$%.2f\n", row, iAmt, pAmt, pRem);
        }
    }

    /**
     * Problem 7. License plate game (40 Points)
     */
    public enum PlayerState {NotInDictionary, LetterMismatch, Good}
    private static void prob7() {
        /*
         * A fun game to play with license plates on a long trip is to come up with
         * the shortest word that uses all the letters in a plate number in order.
         * For example, if a license plate has the letters “FRE”, “free” and “fret” both tie for the best word.
         * If 2 people are playing and one answers “friend” and the other answers “fresh”, “fresh” wins since it is shorter.
         * If there is a 3rd player who tries to win by answering “fre” he won’t win since “fre” is not a valid English word.
         * The letters do not need to be adjacent, but they must be in correct order.
         * So, the following is correct: “deFRayEd”.
         * But, “reFeR” and “reF” are not correct (the capitalized letters are the needed letters in order; note that for these we never get to the E at the end).
         *
         * Your program will do the following in order.
         * •    Ask the user to enter the letters on the license plate.
         * •    Ask player 1 to enter their word.
         * •    Ask player 2 to enter their word. (We’re trusting player 2 to not just copy player 1’s word.)
         * •    Give player 1 one of the following responses as appropriate:
         *      o   “Player 1, your word is not in the dictionary.”
         *      o   “Player 1, your word does not have the needed letters in the right order.”
         *      o   “Player 1, that’s a good word. Its length is N letters.”
         *             Fill in the appropriate value for N.
         * •    Give player 2 an appropriate response, similar to the above.
         * •    Declare who the winner is, or that there is a tie.
         *      o   If neither player enters a “good word”, it is a tie.
         *      o   If only one player enters a “good word”, they’re the winner.
         *      o   If both players enter a “good word”, the shorter one wins.
         *             There is a tie if they’re the both the same length.
         *
         * Important notes
         * •    Case is unimportant. It may be helpful to convert everything the user enters to lowercase.
         * •    All inputs are correctly entered.
         *      o   Nobody will try to confuse your program with spaces, numbers, etc. Only letters will be entered.
         *      o   For the license plate, the user will enter between 1 and 6 letters inclusive.
         *      o   Users enter at least 1 letter for their words.
         * •    Use the given dictionary in dictionary.txt, where each valid word appears on a separate line in lowercase in alphabetical order.
         *      This is available on the website.
         */
        final Dictionary d = new Dictionary(new SortedArrayList<String>());
        d.load(new File("dictionary.txt"));

        // •    Ask the user to enter the letters on the license plate.
        final String plateLetters = inputLine("Enter license plate letters: ").toLowerCase();
        assert 1 <= plateLetters.length() && plateLetters.length() <= 6 : "1-6 letters required"; // competitors don't need to check; not a complete check

        // •    Ask player 1 to enter their word.
        // •    Ask player 2 to enter their word. (We’re trusting player 2 to not just copy player 1’s word.)
        final String[] words = new String[2];
        for (int i = 0; i < words.length; ++i)
            words[i] = inputLine("Player " + (i+1) + ", enter your word: ").toLowerCase();

        // •    Give player 1 one of the following responses as appropriate:
        //      o   “Player 1, your word is not in the dictionary.”
        //      o   “Player 1, your word does not have the needed letters in the right order.”
        //      o   “Player 1, that’s a good word. Its length is N letters.”
        //             Fill in the appropriate value for N.
        // •    Give player 2 an appropriate response, similar to the above.
        final PlayerState[] playerState = new PlayerState[words.length];
        for (int i = 0; i < words.length; ++i) {
            playerState[i] = d.contains(words[i]) ?
                    (wordMatches(plateLetters,words[i]) ? PlayerState.Good : PlayerState.LetterMismatch) :
                    PlayerState.NotInDictionary;
            switch (playerState[i]) {
                case NotInDictionary:
                    System.out.printf("Player %d, your word is not in the dictionary.\n", i + 1);
                    break;
                case LetterMismatch:
                    System.out.printf("Player %d, your word does not have the needed letters in the right order.\n", i+1);
                    break;
                case Good:
                    System.out.printf("Player %d, that’s a good word. Its length is %d letters.\n", i+1, words[i].length());
                    break;
            }
        }
        // •    Declare who the winner is, or that there is a tie.
        //       *      o   If neither player enters a “good word”, it is a tie.
        //      o   If only one player enters a “good word”, they’re the winner.
        //       *      o   If both players enter a “good word”, the shorter one wins.
        //             There is a tie if they’re the both the same length.

        final boolean good[] = new boolean[words.length];
        boolean allGood = true, anyGood = false;
        for (int i = 0; i < words.length; ++i) {
            good[i] = playerState[i] == PlayerState.Good;
            allGood &= good[i];
            anyGood |= good[i];
        }

        assert words.length == 2 : "Logic only implemented for 2 players";
        if (!anyGood) {
            System.out.println("Tie since neither player entered a good word.\n");
        } else if (!allGood) {
            System.out.printf("Player %d wins since they entered the only good word.\n", good[0] ? 1 : 2);
        } else { // allGood, find max
            int winner = (words[0].length() < words[1].length()) ? 1 :
                          ((words[1].length() < words[0].length()) ? 2 : 0);
            if (winner == 0) {
                System.out.println("Tie since both words are good and of equal length.");
            } else {
                System.out.printf("Player %d won since both words were valid but theirs was shorter.\n", winner);
            }

        }

    }
    private static boolean wordMatches(String plate, String word) {
        // ASSUME: all inputs are in lowercase (or all uppercase)
        int plateIdx = 0;
        for (int wordIdx = 0; wordIdx < word.length() && plateIdx < plate.length(); ++wordIdx) {
            if (word.charAt(wordIdx) == plate.charAt(plateIdx)) {
                ++plateIdx;
            }
        }
        return plateIdx == plate.length();
    }

    /**
     * Problem 8. Gray code (40 Points)
     */
    private static void prob8() {
        /* A Gray code is an ordering of patterns of 1s and 0s such that all possible patterns are used but only one bit (1 or 0) changes between each code word.
         * For example, a Gray code of length 2 is: 00, 01, 11, 10.
         * But, 00, 01, 10, 11 is not a Gray code since the code words 01 and 10, which differ in 2 bits, are adjacent in the code.
         * Note also that the last element in a Gray code must be only 1 bit different than the 1st element.
         * You will implement the following algorithm that generates a Gray code for any given length, N, input by the user:
         *   •  Start with all 0s
         *   •  Repeat the following 2N-1 times
         *      o   To generate the next code word, change the rightmost (least significant) bit that does not result in a code word that has already been seen.
         * Your program will input the length from the user and then output the specified sequence with one code word per line. For example:
         *   Number of bits? 3
         *   000 001 011 010 110 111 101 100
         */
        final int N = inputUnsignedInt("Number of bits? ");
        boolean[][] code = new boolean[1<<N][N];
        code[0] = new boolean[N]; // default false
        for(int word = 1; word < 1<<N; ++word) {
           code[word] = code[word-1].clone();
           for(int tryToggle = 0; tryToggle < N; ++tryToggle) {
               code[word][tryToggle] = !code[word][tryToggle]; // tentatively toggle
               boolean match = false;
               for (int row = 0; row < word && !match; ++row) { // row to compare with
                   boolean rowMatch = true;
                   for (int col = 0; col < N; ++col) {
                       rowMatch &= code[word][col] == code[row][col];
                   }
                   match = rowMatch;
               }
               if (match) {
                   code[word][tryToggle] = !code[word][tryToggle]; // undo toggle
                   // note: we should really cover the case when we complete the loop and all N cases matched to catch logic errors [check for final case right here]
               } else {
                   break; // end for since we found the LSB that doesn't duplicate a pattern
               }
           }
        }
        for(int word = 0; word < 1<<N; ++word) {
            for(int bit = N-1; bit >= 0; --bit) {
                System.out.printf("%d", code[word][bit] ? 1 : 0);
            }
            System.out.println();
        }

    }

    /**
     * Problem 9. The friend of my friend is my enemy (40 Points)
     */
    private static void prob9() {
        /* In an ideal world, everybody would be friends, or at least all our friends’ friends would also be our friends.
         * But people, or at least their governments, don’t work this way.
         * Take Iran and Israel, for example, which have antagonistic relations, although each is friendly with Russia.
         * The following graph from the September 13, 2014 issue of The Economist illustrates
         *  the quality of bilateral relations among 14 governments, organizations, and other blocs. [see problem handout]
         *
         * Your job is to read a file representing relations among up to 15 groups, prompt the user for a group number,
         *   and then report any and all groups that are unfriendly towards that group but are among its friends’ friends
         *   (but don’t consider friends’ friends’ friends, etc.).
         * The file will be formatted as follows; you may assume that it is correctly formatted:
         * •    The first line contains an integer, N, from 3 to 15 inclusive that indicates how many groups there are.
         *      This document references the groups using the numbers 0 through N–1.
         * •    Then, there are lines for groups 1 (not 0) through N-1. Let i be the line number.
         * •    Line i contains i integers indicating relations with groups 0 through i-1.
         * •    On a line, integers are separated with one or more spaces or tabs.
         * •    An integer of 1 (or +1) indicates Friendly, 0 indicates Neutral/mistrust, and -1 indicates Unfriendly/enemy.
         * •    In the above chart, N=14, Russia is 0, the Syrian government is 1, al-Qaeda/Jabhat al-Nusra is 12, and the Islamic State is 13.
         * The website contains a sample file representing the above chart.
         */
        final int[][] relationships = readArray(inputLine("Filename? "));
        final int N = relationships.length;
        System.out.printf("%d groups were found.\n", N);
        final int analyze = inputUnsignedInt("Group to analyze? "); // bounds checking not required
        // First, find friends:
        LinkedList<Integer> friends = new LinkedList<>();
        System.out.print("List of friends:"); // Solution is NOT required to report this
        for (int col = 0; col < N; ++col) {
            if (relationships[analyze][col] == 1) { // self is 0, correctly ignored
                friends.push(col);
                System.out.printf(" %d", col);
            }
        }
        System.out.println();

        // Now, find friends' friends (must be a path of length EXACTLY 2)
        final boolean[] ff = new boolean[N]; // default false
        for(int fi : friends) {
            for (int col = 0; col < N; ++col) {
                if (relationships[col][fi] == 1) {
                    ff[col] = true;
                }
            }
        }
        System.out.print("List of friends' friends:"); // Solution is NOT required to report this
        for(int col = 0; col < N; ++col)
            if(ff[col]) System.out.printf(" %d", col);
        System.out.println();

        // Now, find which of those (ff) are enemies
        LinkedList<Integer> ffe = new LinkedList<>();
        for (int pe = 0; pe < N; ++pe) {
            if (ff[pe] && relationships[analyze][pe] == -1) {
                ffe.push(pe);
            }
        }

        System.out.printf("There are %d friends' friends who are enemies of %d and they are:", ffe.size(), analyze);
        for(Integer e : ffe) System.out.printf(" %d", e);
        System.out.println();
    }

    /**Read a NxN array from a file. File provides lower triangular portion.
     * Upper triangular portion is symmetric; diagonal is 0.
     * Assumes properly formatted file.
     * @param fn the name of the file to read from
     * @return the NxN array (N is read from first line of file)
     */
    private static int[][] readArray(String fn) {
        Scanner sc = null;
        int[][] result = null;
        try {
            sc = new Scanner(new File(fn));
        } catch (FileNotFoundException e) {
            System.out.printf("File not found: [%s].\n", fn);
        }
        if (sc != null) {
            final int N = sc.nextInt();

            result = new int[N][N]; // diagonal correctly defaults to 0

            boolean success = true;

            for(int row = 1; row < N && success; ++row) { // 1 since nothing to read in 0 due to LU/symmetry
                for(int col = 0; col < row; ++col) {
                    if (sc.hasNextInt()) {
                        result[row][col] = sc.nextInt();
                    }
                    else {
                        success = false;
                        break;
                    }
                }
            }

            if(!success) {
                System.out.printf("Fewer than the expected number of elements found in [%s].\n", fn);
            }

            if(success) { // complete upper half
                for (int row = 0; row < N; ++row) { // non-degenerate bound is < N-1
                    for (int col = row+1; col < N; ++col) {
                        result[row][col] = result[col][row];
                    }
                }
            }

            sc.close();
        }
        return result;
    }

    /**
     * prompt for int, skipping any invalid preamble
     * @param prompt prompt to display
     * @return the entered int
     */
    private static int inputInt(String prompt)
    {
        System.out.print(prompt);
        while (!sc.hasNextInt())
            sc.next(); // discard non-int token
        int retVal = sc.nextInt();
        sc.nextLine(); // dump CR from stream
        return retVal;
    }
    /**
     * prompt for unsigned int, skipping any invalid preamble
     * @param prompt prompt to display
     * @return the entered int
     */
    private static int inputUnsignedInt(String prompt)
    {
        System.out.print(prompt);
        int result = -1;
        while (result < 0)
        {
            while (!sc.hasNextInt())
                sc.next(); // discard non-int token
            result = sc.nextInt();
        }
        sc.nextLine(); // dump CR from stream
        return result;
    }

    /**
     * prompt for a double, skipping any invalid preamble
     * @param prompt prompt to display
     * @return the entered double
     */
    private static double inputDouble(String prompt)
    {
        System.out.print(prompt);
        while (!sc.hasNextDouble())
            sc.next(); // discard non-matching token
        double retVal = sc.nextDouble();
        sc.nextLine(); // dump CR from stream
        return retVal;
    }

    private static String inputLine(String prompt)
    // Simple prompt for a line of text
    {
        System.out.print(prompt);
        String result;
        // Key here is to discard empty lines (including \n's left in the buffer), but we make it a bit more robust to whitespace...
        do {
            result = sc.nextLine();
            result = result.trim();
        } while (result.length() == 0);
        return result;
    }

    private final static Scanner sc = new Scanner(System.in);
}
