//package lab6;
/**
 * SinglyLinkedList.java
 * This is a partial implementation of a SortedArrayList class.
 * The class extends ArrayList (which provides most of the
 * functionality) and rewrites the implementation of a select
 * set of methods to ensure that elements within the sorted
 * array are organized in ascending order.  Only a few of the
 * methods have been optimized for a sorted array.  The methods
 * that have been optimized are explicitly overridden in
 * this subclass.
 *
 * The following methods have been overridden as unsupported since
 * they do not make sense in the context of a sorted container:
 * 	public E set(int arg0, E arg1);
 *  public void add(int arg0, E arg1);
 *
 * Original written: 12-9-2005
 * @author t a y l o r@msoe.edu
 *
 * @param <E> The type of data stored in the list.
 */
import java.util.List;
import java.util.ArrayList;
import java.util.Collections;
//public class SortedArrayList<E extends Comparable<E>> extends ArrayList<E> implements List<E>

public class SortedArrayList<E extends Comparable<E>> extends ArrayList<E> implements List<E>
{

    /**
     * Automatically generated (via Eclipse) serialVersionUID.
     */
    private static final long serialVersionUID = 7373358239830027989L;

    /**
     * Returns <tt>true</tt> if this list contains the specified element.
     * More formally, returns <tt>true</tt> if and only if this list contains
     * at least one element <tt>e</tt> such that
     * <tt>(o==null&nbsp;?&nbsp;e==null&nbsp;:&nbsp;o.equals(e))</tt>.
     * Makes use of a binarySearch to achieve O(log n) execution time.
     *
     * @param o element whose presence in this list is to be tested.
     * @return <tt>true</tt> if this list contains the specified element.
     * @throws ClassCastException if the type of the specified element
     * 	       is incompatible with this list (optional).
     * @throws NullPointerException if the specified element is null and this
     *         list does not support null elements (optional).
     */
    @SuppressWarnings("unchecked")
    @Override
    //public boolean contains(E arg0) // original
    public boolean contains(Object arg0)
    {
        int index = Collections.binarySearch(this, (E)arg0); // added cast
        return index>=0;
    }

    /**
     * Adds the specified element to the list while maintaining
     * sorted order. <p>
     *
     * Uses binarySearch to find the location to insert the element
     * and then calls add(int, E) to do the inserting.
     *
     * @param o element to be added to this list.
     * @return <tt>true</tt> (as per the general contract of the
     *            <tt>Collection.add</tt> method).
     *
     * @throws ClassCastException if the class of the specified element
     * 		  prevents it from being added to this list.
     * @throws IllegalArgumentException if some aspect of this element
     *            prevents it from being added to this list.
     */
    @SuppressWarnings("unchecked")
    public boolean add(E arg0) {
        int index = Collections.binarySearch(this, arg0);
        if (index<0)
        {
            ++index;
            index *= -1;
        }
        super.add(index, arg0);
        return true;
    }

    /**
     * @throws UnsupportedOperationException so that a client cannot
     *        cannot corrupt the sorted order of the list.
     */
    public E set(int arg0, E arg1) {
        throw new UnsupportedOperationException();
    }

    /**
     * @throws UnsupportedOperationException so that a client cannot
     *        cannot corrupt the sorted order of the list.
     */
    public void add(int arg0, E arg1) {
        throw new UnsupportedOperationException();
    }

    /**
     * Returns the index in this list of the first occurrence of the specified
     * element, or -1 if this list does not contain this element.
     * More formally, returns the lowest index <tt>i</tt> such that
     * <tt>(o==null ? get(i)==null : o.equals(get(i)))</tt>,
     * or -1 if there is no such index. <p>
     *
     * Makes use of binarySearch to find the index.  If the item is not
     * found, the index is set to -1 to match the interface specification.
     *
     * @param o element to search for.
     * @return the index in this list of the first occurrence of the specified
     * 	       element, or -1 if this list does not contain this element.
     */
    @SuppressWarnings("unchecked")
    public int indexOf(E arg0) {
        int index = Collections.binarySearch(this, arg0);
        if(index<0)
        {
            index = -1;
        }
        return index;
    }
}
